function GDT = GDTopt(Yref,A_n,B_n,N,bet,plantfunstr,X0,Y0,DELTAU,P1,P2,P3,P4,P5,P6)
% GDTOPT    Genealogical decision tree - optimization procedure
%
%   GDT = GDTopt(Yref,A_n,B_n,N,bet,plantfunstr,X0,Y0,deltaU,P1,P2,P3,P4,P5,P6)
%
% Inputs:
% -------
% Yref          target trajectory
% A_n, B_n      tuning matrices
% N, bet        # particles, tuning parameter
% plantfunstr   name of the function that is used for simulating the plant
% X0, Y0        initial state and initial output of the plant
% deltaU        a flag indicating optimization of incremental control sequence (deltaU=1) or absolute control
% P1,P2,P3,...  additional arguments to be passed to plant simulation function
%               The plant must be given by a function
%                   [yk1, xk1] = fun(xk, uk, P1, P2, P3, ...)
%               where 
%               - yk1 and xk1 are the output and state vectors at instant k+1
%               - xk and uk are the state and control vectors at instant k
%               - P1, P2, ... are additional arguments, 
%               For example, in RTP-plant model (rtpmod.m) they are
%                   P1 - sampling time
%                   P2 - measurement matrix
%
%
% GDT-structure
% -------------
% Information of particles is stored in a structure named GDT.
% GDT(i) is the structure for the i'th particle, with fields:
%   U_hist      past plant control sequence
%   X_hist      past plant state sequence
%   Y_hist      past plant output sequence
%   J_hist      past cost history
%   parent_hist past indexes to parent particles
%   X           current plant state
%   Y           current plant output
%   J_T         current total cost
%   Jy          current deviation cost
%   Ju          current control cost
%   parent      parent of current particle
%   p_un        unnormalized resampling weight of particle
%   survived    number of survived particles in the population at each resampling
%   Yref, A_n, B_n, bet, N, plantfunstr, X0, Y0, DELTAU     as above

% Enso Ikonen 4/2006 University of Oulu

Nargin = nargin;
if Nargin<9
    % assume that first argument is a GDT structure
    % read parameters from there 
    GDT = Yref;
    Yref = GDT(1).Yref;
    A_n = GDT(1).A_n; B_n = GDT(1).B_n; 
    N = GDT(1).N; bet = GDT(1).bet; 
    plantfunstr = GDT(1).plantfunstr;
    X0 = GDT(1).X0; Y0 = GDT(1).Y0;
    DELTAU = GDT(1).DELTAU;    % type of search    
    % pass additional arguments to plantfunstr
    switch(Nargin)
    case 2, P1=A_n
    case 3, P1=A_n;P2=B_n;
    case 4, P1=A_n;P2=B_n;P3=N;
    case 5, P1=A_n;P2=B_n;P3=N; P4=bet;
    case 6, P1=A_n;P2=B_n;P3=N; P4=bet; P5=plantfunstr;
    case 7, P1=A_n;P2=B_n;P3=N; P4=bet; P5=plantfunstr; P6 = X0;
    case 8, P1=A_n;P2=B_n;P3=N; P4=bet; P5=plantfunstr; P6 = X0; P7 = Y0;
    end
    Nargin = Nargin+8;
end

% Set optimization horizon
T = size(Yref,1);
NOLLA = 0;

% store parameters to GDT structure
clear GDT
for i=1:N
    GDT(i).Yref = Yref;
    GDT(i).A_n = A_n; GDT(i).B_n = B_n; 
    GDT(i).N = N; GDT(i).bet = bet; 
    GDT(i).plantfunstr = plantfunstr;
    GDT(i).X0 = X0; GDT(i).Y0 = Y0;
    GDT(i).DELTAU = DELTAU;    % type of search   
end

% OPTIMIZE
% --------
t0 = clock; TC00=clock;     % initialize timers (for examination of code performance, only)
LOGALL=0;   % set initial flags

%DELTAU=1        % set flag for incremental control
                % DELTAU = 0    optimize absolute values
                % DELTAU = 1    optimize incremental control values
n_I = length(A_n);  % # inputs
n_O = length(B_n);  % # outputs

% optimization loop
for n=1:T   % loop target sequence from the begenning to end of horizon
    TC0 = clock;    % set timer 
    yref = Yref(n,:)';  % target at step n
    for i=1:N   % loop particles
        % if this is the first step, initialize particles
        if n==1
            GDT(i).U_hist=[];   % null history
            GDT(i).X_hist=[];
            GDT(i).Y_hist=[];
            GDT(i).J_hist=[];
            GDT(i).parent_hist=[];
            GDT(i).X = X0;  % plant state
            GDT(i).Y = Y0;  % plant output
            GDT(i).J_T = 0; % cost
            GDT(i).parent = i;    % parent
        end
        
        % generate random inputs U_n^i
        %GDT(i).U = sqrt(diag(A_n)).*randn(n_I,1);
        GDT(i).U = Urand(A_n);
        if NOLLA, if i==1, GDT(1).U = zeros(size(Urand(A_n))); end; end
        
        % evaluate plant output Y_n^i and costs J_n^i
        GDT(i).Jy = (GDT(i).Y-yref)'*inv(B_n)*(GDT(i).Y-yref);
        GDT(i).Ju = GDT(i).U'*inv(A_n)*GDT(i).U;
        GDT(i).J_T = GDT(i).Jy + GDT(i).Ju + GDT(i).J_T;
        
        % Set resampling weight (unnormalized p)
        GDT(i).p_un =  exp(-(1/2)*bet*GDT(i).Jy);
        
        % collect all inputs for this ancestral line
        GDT(i).U_hist(n,1:n_I) = GDT(i).U'; % add to past history
        GDT(i).X_hist(n,:) = GDT(i).X';
        GDT(i).Y_hist(n,:) = GDT(i).Y';
        GDT(i).J_hist(n,:) = [GDT(i).J_T GDT(i).Jy GDT(i).Ju];
        GDT(i).parent_hist(n,1) = GDT(i).parent;
        
        % Eval plant (moved)
    end
    TC1 = clock;    % timer
    
    % Normalize resampling probabilities
    p = [GDT.p_un]; 
    if sum(p)==0, p=(1/N)*ones(1,N); else; p = p/sum(p); end
    TC2 = clock;    % timer
    
    % Resample
    if 0
        % implementation or resampling as suggested by Najim and DelMoral
        J = UGener(p);
    else
        % simple implementation of resampling
        for j=1:N
            J(j,1) = min(find(cumsum(p)>rand));
        end
    end
    if NOLLA, J(1)=1; end
    TC3 = clock;    % timer
    
    % Evaluate plant at n+1 for all inputs that survived 
    % - make list K of survived lines
    k = 0; K=[];
    for j=1:N,
        if any(J==j), k = k+1; K(k) = j; end
    end
    % - simulate plant for all lines in K
    if 0
        % - batch -
    else
        % simulate plant for each particle at a time (one-by-one)
        for k=1:length(K)
            % Set index for an ancestral line that survived
            i = K(k);
            % Evaluate plant for next n
            if DELTAU
                % Incremental control
                %Unom = P3; 
                %if ~exist('P4'); IN=1; else, IN = P4; end
                if n==1
                    % first control is nominal control + delta-u
                    %GDT(i).U_hist = GDT(i).U_hist + Unom(IN)'; 
                    GDT(i).U_hist = GDT(i).U_hist; 
                    U_in = GDT(i).U_hist';
                else
                    % after 1st, controls are summed
                    U_in = sum(GDT(i).U_hist)';
                end
            else
                % absolute control
                U_in = GDT(i).U;
            end        
            % plot something for the impatient...
            if n==1
                if k==1; te0 = clock; end
                if k<10, fprintf('x'); end; 
                if k==10, 
                    te1 = etime(clock,te0)/10;
                    fprintf('...(%s/n, %s/T)\n',sec2time(N*te1),sec2time(N*T*te1)); 
                end
            end

            % evaluate plantfunstr-command with a given number of arguments
            switch(Nargin-9)
            case 0
                [GDT(i).Y,GDT(i).X] = feval(plantfunstr,GDT(i).X,U_in);
                %Xk = GDT(i).X;
                %Uk = U_in;
                %[GDT(i).Y,GDT(i).X] = eval(plantfunstr);
            case 1
                [GDT(i).Y,GDT(i).X] = feval(plantfunstr,GDT(i).X,U_in,P1);
            case 2
                [GDT(i).Y,GDT(i).X] = feval(plantfunstr,GDT(i).X,U_in,P1,P2);
            case 3
                [GDT(i).Y,GDT(i).X] = feval(plantfunstr,GDT(i).X,U_in,P1,P2,P3);
            case 4
                [GDT(i).Y,GDT(i).X] = feval(plantfunstr,GDT(i).X,U_in,P1,P2,P3,P4);
            case 5
                [GDT(i).Y,GDT(i).X] = feval(plantfunstr,GDT(i).X,U_in,P1,P2,P3,P4,P5);
            case 6
                [GDT(i).Y,GDT(i).X] = feval(plantfunstr,GDT(i).X,U_in,P1,P2,P3,P4,P5,P6);
            otherwise
                error('Not implemented for more than 6 arguments')
            end
            
        end
    end
    TC4 = clock;    % timer
    
    % Death & birth
    I = [1:N]';
    GDT(I) = GDT(J);
    
    % log parent for each particle
    SH = sum(sign(diff(sort([GDT.parent_hist]'))))+1;
    for j=1:N
        GDT(j).parent = J(j);
        GDT(j).survived(n) = SH(n);
    end
    
    % find optimal trajectory (index i_opt)
    %J_T = [GDT(I).J_T]; % costs for all trajectories
    %i_opt = find( J_T == min(J_T) ); i_opt = i_opt(1);  % find the smallest

    TC5 = clock;    % timer
    % Plot current solution (plot whatever is necessary...)
    % display progress and code efficiency
    if (n>1 & (n==T | etime(clock,t0)>3)) | n<5
        tn = etime(TC5,TC00)/n; % seconds/iteration
        tf = (T-n)*tn;          % estimated time at the end of simulation
        nd = sum((diff(sort(J))~=0))+1;
        %disp(sprintf('%1.2g s/iteration, %s to go',tn,sec2time(tf)));
        % plot
        GDTplot(GDT);
        % show speed in window title
        str = get(gcf,'Name');
        if n<T
            % show time to go
            set(gcf,'Name',sprintf('(%s) %s',sec2time(tf),str));
        else
            % show total elapsed time
            set(gcf,'Name',sprintf('%s (%s)',str,sec2time(etime(clock,TC00))));
        end
        shg; drawnow;
        t0 = clock;
    end

end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [s,t]=sec2time(sec,t0);
%SEC2TIME	Convert seconds to secs,mins,hours etc...
%
%	[s,t] = sec2time(sec)
%	[s,t] = sec2time(sec,t0)	add sec seconds to t0
%
%s	time/date string
%t	time vector [yy mm dd hh mm ss]
%
%Does not work over month boundaries!

if nargin==2
	t00=clock;
    t00=[t00(1:2) 0 0 0 0];  % start from the first day of the month
	sec0=etime(t0,t00); % seconds elapsed between t00 to t0
	sec=sec0+sec;       % add sec seconds
end

t=zeros(1,6);
% Convert from seconds to secs, mins, hours, etc.
bases=cumprod([1 60 60 24 30 12]);
bases=bases(6:-1:1);
%keyboard
for i=3:6
	t(i)=fix(sec/bases(i));
	sec=sec-t(i)*bases(i);
end
if nargin>1,t(1:2)=t0(1:2);end

% Create string
for i=1:6, if (t(i)~=0), break; end, end
if i==6, T=t(6);
elseif i==5, T=t(5:6);
elseif i==4, T=t(4:6);
else, T=[t(4:6) t(3) t(2) t(1)];
end
s=sprintf('%2d:%02d:%02d %d/%d/%d-',T);
s=s(1:length(s)-1);
